-- by mor2000

--------------------
-- DisplayTrigger (shows sth in the top left hud)

UPK_DisplayTrigger={}
local UPK_DisplayTrigger_mt = Class(UPK_DisplayTrigger, UniversalProcessKit)
InitObjectClass(UPK_DisplayTrigger, "UPK_DisplayTrigger")
UniversalProcessKit.addModule("displaytrigger",UPK_DisplayTrigger)

function UPK_DisplayTrigger:new(isServer, isClient, customMt)
	if customMt == nil then
		customMt = UPK_DisplayTrigger_mt
	end
	local self = UniversalProcessKit:new(isServer, isClient, customMt)
	registerObjectClassName(self, "UPK_DisplayTrigger")
	self.getShowInfo=GreenhousePlaceable.getShowInfo
	return self
end

function UPK_DisplayTrigger:load(id, parent)
	UPK_DisplayTrigger:superClass().load(self, id, parent)
	
	table.insert(self.triggerIds,id)
	addTrigger(id, "triggerCallback", self)
	self:register(true)
	
	self.playerInRange=false
	self.vehiclesInRange={}
	
    local function tobool(val)
  		return not (val == nil or val == false or val == 0 or val == "0" or val == "false" )
    end
	
	self.onlyFilled = tobool(getUserAttribute(id, "onlyFilled"))
	self.showFillLevel = tobool(Utils.getNoNil(getUserAttribute(id, "showFillLevel"),true))
	self.showPercentage = tobool(getUserAttribute(id, "showPercentage"))
	
	return true
end

function UPK_DisplayTrigger:delete()
	UPK_DisplayTrigger:superClass().delete(self)
end

function UPK_DisplayTrigger:update(dt)
	if self:getShowInfo() then
		local fluid_unit_short=g_i18n:getText("fluid_unit_short")
		for k,v in pairs(self.acceptedFillTypes) do
			if v then
				local fillLevel=self:getFillLevel(k)
				if fillLevel>0 or not self.onlyFilled then
					local text=g_i18n:getText(UniversalProcessKit.fillTypeIntToName[k]) .. ": "
					if self.showFillLevel then
						text=text.." "..math.max(0,math.ceil(fillLevel)) .. "[" .. fluid_unit_short .. "]"
					end
					if self.showPercentage then
						text=text.." "..math.max(0,math.ceil(fillLevel/self.capacity*100)) .. "%"
					end
	    			g_currentMission:addExtraPrintText(text)
				end
			end
		end
	end
end

function UPK_DisplayTrigger:triggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
	-- use these numbers for the collisionMask (or a combination/ sum of it)
	-- http://gdn.giants-software.com/thread.php?categoryId=16&threadId=677
	-- player = 2^20 = 1048576
	-- tractors = 2^21 = 2097152
	-- combines = 2^22 = 4194304
	-- fillables = 2^23 = 8388608
	-- all = 15728640
	
	if self.isEnabled and (onEnter or onLeave) then
		if g_currentMission.player ~= nil and otherId == g_currentMission.player.rootNode then
			self.playerInRange = onEnter==true
		else
			local vehicle = g_currentMission.nodeToVehicle[otherId]
			if vehicle ~= nil then
				--print("DisplayTrigger: "..tostring(onEnter))
				if onEnter then
					self.vehiclesInRange[vehicle] = true
				else
					self.vehiclesInRange[vehicle] = nil
				end
			end
		end
	end
end