-- by mor2000

UniversalProcessKit = {}
local UniversalProcessKit_mt = Class(UniversalProcessKit, Object)
InitObjectClass(UniversalProcessKit, "UniversalProcessKit")

function UniversalProcessKit:new(isServer, isClient, customMt)
	if customMt == nil then
		customMt = UniversalProcessKit_mt
	end
	local self = Object:new(isServer, isClient, customMt)
	registerObjectClassName(self, "UniversalProcessKit");
	
	self.rootNode = 0
	self.triggerId = 0
	self.nodeId = 0
	
	self.parent=nil
	self.kids={}
	self.type=nil
	
	self.fillType=Fillable.FILLTYPE_UNKNOWN
	
	--self.updateFillTypesDirtyFlag=self:getNextDirtyFlag()
	
	-- lets make things easy dealing with fillLevels, parents and stuff
	self.fillLevels={}
	setmetatable(self.fillLevels, {
		__index = function(t,k)
			if k==UniversalProcessKit.FILLTYPE_MONEY then
				return g_currentMission:getTotalMoney()
			end
			if rawget(t,k)==nil then
				if type(rawget(self,"parent"))=="table" then
					return self.parent.fillLevels[k]
				end
				return nil
			end
			return t[k]
		end,
		__newindex=function(t,k,v)
			if type(rawget(self,"parent"))=="table" then
				self.parent.fillLevels[k]=v
			else
				rawset(t,k,v)
			end
		end,
		__call=function(func,...)
			local t={}
			local args=...
			if type(args)~="table" then
				args={...}
			end
			for k,v in pairs(args) do
				local type=type(v)
				if type=="number" then
					if self.fillLevels[v]~=nil then
						table.insert(t,self.fillLevels[v])
					end
				end
			end
			return __c(t)
		end	
	})
	
	self.acceptedFillTypes={}
	setmetatable(self.acceptedFillTypes, {
		__index = function(t,k)
			if rawget(t,k)==nil then
				if type(rawget(self,"parent"))=="table" then
					return self.parent.acceptedFillTypes[k]
				end
				return nil
			end
			return t[k]
		end,
		__newindex=function(t,k,v)
			if type(rawget(self,"parent"))=="table" then
				self.parent.acceptedFillTypes[k]=v
			else
				rawset(t,k,v)
			end
		end,
		__len=function(t)
			local len=0
			for i=1,128 do
				if self.acceptedFillTypes[i]~=nil then
					len=len+1
				end
			end
		end,
		__call=function(func,...)
			local t={}
			local args=...
			if type(args)~="table" then
				args={...}
			end
			for k,v in pairs(args) do
				local type=type(v)
				if type=="number" then
					if self.acceptedFillTypes[v]~=nil then
						table.insert(t,self.acceptedFillTypes[v])
					end
				end
			end
			return __c(t)
		end	
	})
	
	return self
end

function UniversalProcessKit:load(id,parent)
	self.rootNode = id
	self.triggerId = id
	self.nodeId = id
	
	self.parent = parent
	
	self.x,self.y,self.z = getTranslation(self.nodeId)
	self.pos = __c({self.x,self.y,self.z})
	self.wpos = __c({getWorldTranslation(self.nodeId)})
	self.rot = __c({getRotation(self.nodeId)})
	
	self.triggerIds={}
	
	local function gmatch(str, pattern)
		local arr={}
		if type(str)=="string" then
			for v in string.gmatch(str,pattern) do
				table.insert(arr,v)
			end
		end
		return arr
	end
	
    local function tobool(val)
  		return not (val == nil or val == false or val == 0 or val == "0" or val == "false" )
    end
	
	

	-- accepted fillTypes or fillTypes to store
	local acceptedFillTypesString = getUserAttribute(self.nodeId, "fillTypes")
	for _,v in pairs(UniversalProcessKit.fillTypeNameToInt(gmatch(acceptedFillTypesString, "%S+"))) do
		rawset(self.acceptedFillTypes,v,true)
	end
	
	-- capacity
	self.capacity = tonumber(getUserAttribute(self.nodeId, "capacity"))
	if self.capacity==nil and self.parent~=nil then
		self.capacity=self.parent.capacity	-- capacity cannot be changed on the run
	end
	if self.capacity==nil then
		self.capacity=math.huge
	end

	-- loading kids (according to known types of modules)
	-- kids are loading their kids and so on..
	
	self.kids={}
	self:findChildren(id,1)
	
	-- enable processing of stuff
	
	self.name = getName(self.nodeId)
	self.type = getUserAttribute(self.nodeId, "type")
	self.isEnabled = tobool(Utils.getNoNil(getUserAttribute(self.nodeId, "isEnabled"), "true"))

	-- MapHotspot
	
	self.MapHotspotName = getUserAttribute(self.nodeId, "MapHotspot")
	local mapHotspotIcons = { TipPlace = "dataS2/missions/hud_pda_spot_tipPlaceGold.png" }
	if self.MapHotspotName~=nil and mapHotspotIcons[self.MapHotspotName]~=nil then
		self.MapHotspotIcon = mapHotspotIcons[self.MapHotspotName]
	else
		self.MapHotspotIcon = mapHotspotIcons["TipPlace"]
	end

	if self.MapHotspotName~=nil then
		self:showMapHotspot(self.appearsOnPDA)
	end

	self:register()

	return true
end

function UniversalProcessKit:findChildren(id,numKids)
	local numChildren = getNumOfChildren(id)
	for i=1,numChildren do
		local childId = getChildAt(id, i-1)
		if childId~=nil or childId~=0 then
			local scriptCallback = getUserAttribute(childId, "scriptCallback")
			if scriptCallback==nil then
				local type = getUserAttribute(childId, "type")
				local moduleFound=false
				for k,v in pairs(UniversalProcessKit.Modules) do
					if type==v.type then
						self.kids[numKids]=v.class:new(self.isServer,self.isClient)
						if self.kids[numKids]~=nil then
							self.kids[numKids]:load(childId,self)
							self.kids[numKids]:register(true)
							numKids=numKids+1
						end
						moduleFound=true
						break
					end
				end
				if not moduleFound then
					self:findChildren(childId,numKids)
				end
			end
		end
	end
end

function UniversalProcessKit:delete()
	for k,_ in pairs(self.kids) do
		if self.kids[k].delete ~=nil then
			self.kids[k].delete(self.kids[k])
		end
	end
	for k,v in pairs(self.triggerIds) do
		removeTrigger(v)
	end
	if self.mapHotspot ~= nil then
		g_currentMission.missionPDA:deleteMapHotspot(self.mapHotspot)
	end
	self:unregister()
	UniversalProcessKit:superClass().delete(self)
end

function UniversalProcessKit:readStream(streamId, connection)
	UniversalProcessKit:superClass().readStream(self, streamId, connection)
	if connection:getIsServer() then
		if type(rawget(self,"fillLevels"))=="table" then
			for k,v in pairs(self.fillLevels) do
				local fillType = streamReadInt8(streamId)
				self.fillLevels[fillType] = streamReadFloat32(streamId)
			end
		end
	end
end

function UniversalProcessKit:writeStream(streamId, connection)
	UniversalProcessKit:superClass().writeStream(self, streamId, connection)
	if not connection:getIsServer() then
		if type(rawget(self,"fillLevels"))=="table" then
			for k,v in pairs(self.fillLevels) do
				streamWriteInt8(streamId, k)
				streamWriteFloat32(streamId, v)
			end
		end
	end
end

function UniversalProcessKit:readUpdateStream(streamId, timestamp, connection)
	UniversalProcessKit:superClass().readUpdateStream(self, streamId, timestamp, connection)
	self:readStream(streamId, connection)
end

function UniversalProcessKit:writeUpdateStream(streamId, connection, dirtyMask)
	UniversalProcessKit:superClass().writeUpdateStream(self, streamId, connection, dirtyMask)
	self:writeStream(streamId, connection)
end

function UniversalProcessKit:update(dt)
	-- do sth with time (ms)
end

function UniversalProcessKit:getFillType()
	return self.fillType
end

function UniversalProcessKit:setFillType(fillType)
	if fillType~=nil then
		self.fillType = fillType
	else
		self.fillType = Fillable.FILLTYPE_UNKNOWN
	end	
end

function UniversalProcessKit:getFillLevel(fillType)
	if fillType~=nil then
		return self.fillLevels[fillType] or 0
	end
	return nil
end

function UniversalProcessKit:setFillLevel(fillLevel, fillType, raw)
	if fillType~=nil and fillType~="money" then
		if fillLevel==nil then
			if raw then
				rawset(self.fillLevels,fillType,nil)
				-- event
			end
			return nil
		else
			if fillLevel<=self.capacity then
				if not raw then
					self.fillLevels[fillType]=fillLevel
				else
					rawset(self.fillLevels,fillType,fillLevel)
					-- event
				end
				return 0
			else
				if not raw then
					self.fillLevels[fillType]=self.capacity
				else
					rawset(self.fillLevels,fillType,self.capacity)
					-- event
				end
				return fillLevel-self.capacity
			end
		end
	elseif fillType=="money" then
		-- can't set money
	end
	return nil
end

function UniversalProcessKit:addFillLevel(deltaFillLevel, fillType)
	if deltaFillLevel~=nil then
		if fillType==UniversalProcessKit.FILLTYPE_MONEY then
			if deltaFillLevel<0 then
				deltaFillLevel=-math.min(g_currentMission:getTotalMoney(),-deltaFillLevel)
			end
			g_currentMission:addMoney(deltaFillLevel, 1, self.statName or "other")
			return deltaFillLevel
		end
		-- how much of deltaFillLevel was added to the fillLevel?
		return deltaFillLevel-self:setFillLevel(self:getFillLevel(fillType)+deltaFillLevel, fillType)
	end
	return 0
end

-- show or hide an icon on the pda map
function UniversalProcessKit:showMapHotspot(on)
	if on==true and self.mapHotspot == nil then
		local iconSize = g_currentMission.missionPDA.pdaMapWidth / 15
		local x,_,z = unpack(self.pos)
		self.mapHotspot = g_currentMission.missionPDA:createMapHotspot(self.MapHotspotName, self.MapHotspotIcon, x, z, iconSize, iconSize * 1.3333333333333333, false, false, false, 0, true)
	end
	if on==false and self.mapHotspot ~= nil then
		g_currentMission.missionPDA:deleteMapHotspot(self.mapHotspot)
	end
end

function UniversalProcessKit:setEnable(isEnable)
	self.isEnabled=isEnable
end

function UniversalProcessKit:loadFromAttributesAndNodes(xmlFile, key)
	key=key.."."..self.name
	for k,v in pairs(UniversalProcessKit.fillTypeIntToName) do
		local fillLevel = getXMLFloat(xmlFile, key .. "#" .. tostring(v))
		if fillLevel~=nil then
			self:setFillLevel(fillLevel,k,true)
		end
	end
	
	for k,v in pairs(self.kids) do
		v:loadFromAttributesAndNodes(xmlFile, key)
	end

	return self:loadExtraNodes(xmlFile, key)
end

function UniversalProcessKit:getSaveAttributesAndNodes(nodeIdent)
	local attributes=""
	
	local nodes = "\t<"..tostring(self.name)
	
	local extraNodes=""
	for k,v in pairs(UniversalProcessKit.fillTypeIntToName) do
		local fillLevel=rawget(self.fillLevels,k)
		if fillLevel~=nil and fillLevel>=0.001 then
			extraNodes = extraNodes .. " " .. tostring(v) .. "=\"" .. tostring(math.floor(fillLevel*1000+0.5)/1000) .. "\""
		end
	end

	extraNodes=extraNodes..self:getSaveExtraNodes(nodeIdent)
	
	local nodesKids=""
	for k,v in pairs(self.kids) do
		local attributesKid, nodesKid = v:getSaveAttributesAndNodes(nodeIdent)
		attributes = attributes .. attributesKid
		if nodesKid~="" then
			nodesKids = nodesKids .. nodesKid
		end
	end
	
	if nodesKids=="" then
		if extraNodes=="" then
			nodes=""
		else
			nodes = nodes .. extraNodes .. " />\n"
		end
	else
		nodes = nodes .. extraNodes ..">\n" .. string.gsub(nodesKids,"\n","\n\t") .. "\n\t</"..tostring(self.name)..">"
	end
	
	return attributes, nodes
end

-- use this function to load your extra Nodes (YourClass:loadExtraNodes)
function UniversalProcessKit:loadExtraNodes(xmlFile, key)
	return true
end

-- use this function to save your own values (YourClass:getSaveExtraNodes)
function UniversalProcessKit:getSaveExtraNodes(nodeIdent)
	return ""
end

-- to communicate orders between modules
-- possible receivers
-- ALL: all modules and base execute the action (if they have it)
-- ALL_TYPE: all modules of type TYPE execute the action (ie ALL_FILLTRIGGER)
-- name: a specific module with name
-- NOT TESTED, NOT IMPLEMENTED

function UniversalProcessKit:getType()
	return self.type
end

function UniversalProcessKit:getName()
	return self.name
end

function UniversalProcessKit:induceAction(sender, receiverName, transmitter, action, value)
	if receiverName.sub(1,4)=="ALL_" and receiverName.sub(5)==string.upper(self.type) then
		self:induceActionToParent(sender, receiverName, transmitter, action, value)
		self:induceActionToKids(sender, receiverName, transmitter, action, value)
		self:executeAction(action, value)
	elseif receiverName=="ALL" then
		self:induceActionToParent(sender, receiverName, transmitter, action, value)
		self:induceActionToKids(sender, receiverName, transmitter, action, value)
		self:executeAction(action, value)
	elseif receiverName==self.name then
		self:executeAction(sender, action, value)
	end	
end

function UniversalProcessKit:induceActionToParent(sender, receiverName, transmitter, action, value)
	if self.parent~=nil and self.parent~=transmitter then
		self.parent:induceAction(sender, receiverName, self, action, value)
	end
end

function UniversalProcessKit:induceActionToKids(sender, receiverName, transmitter, action, value)
	for i in #self.kids do
		if self.kids[i].getType~=nil and self.kids[i]:getType()~=nil and self.kids[i]~=transmitter then
			self.kids[i]:induceAction(sender, receiverName, self, action, value)
		end
	end
end

-- modify this function in modules to respond to action calls of other modules
function UniversalProcessKit:executeAction(sender, action, value)
	-- yet missing: sendEvent to synchronize - UniversalProcessKitEvent
end
